//==============================================================================
// ContigEvalUtils.hh
//
// ----------------------------------------------------------------------------
// $Id: //CBT-depot/CompBioToolsApps/bac-stitcher/ContigEvalUtils.hh#7 $
//------------------------------------------------------------------------------
// Utilities to aid in finding problems in contigs.  
//
//==============================================================================
//.......1.........2.........3.........4.........5.........6.........7.........8

#include "CompBioTools.hh"


#ifndef _CONTIG_EVALUTILS
#define _CONTIG_EVALUTILS


void FindLeftTail(vector<int> &TCover, int &leftTail);
void FindRightTail(vector<int> &TCover, int &rightTail);


class ContigCheck{
public:

  ContigCheck(){
	 mbMerged = false;
	 mbUsed = false;
  }


  string mContigName;
  string mSide;
  int    mStart;
  int    mEnd;

  // Two checks may be good and close and so will be treated more or less
  // as one point for a time.  Need to record this fact along with the 
  // original coordinates.  mStart and mEnd will be set to something in 
  // the middle of these two coordinates. 
  int    mStartA;
  int    mEndA;
  string mSideA;

  int    mStartB;
  int    mEndB;
  string mSideB;

  int    mPairSize;

  bool   mbMerged;
  int    mConfidence;

  bool   mbUsed; 
};

class ContigChecks : public vector<ContigCheck>{
public:

	void BUG(){
		cout<<"BUG"<<endl;
	}

	friend const bool operator<(const ContigCheck a,const ContigCheck b){
		return(a.mStart < b.mStart);
	}

	int FindClosestMate(int cIdx,int minSize);
  void MergePairs(int msize);

  void GetChecksFrom(string cName,vector<ContigCheck> &Checks){
	 for(int i = 0;i< size();i++){
		if ((*this)[i].mContigName == cName){		  
		  Checks.push_back((*this)[i]);
		}
	 }
  }

  
  void Read(string Name){
	 izipstream fin;
	 fin.open(Name.c_str());

	 if (fin.fail()) {
		cerr<<"Unable to open problem file: "<<Name<<endl;
		exit(1);
	 }

	 string temp;
	 // First two lines are junk
	 GetLine(fin,temp);
	 GetLine(fin,temp);

	 int nextIdx = 0;
	 do{
		GetLine(fin,temp);
		if (!ContainsOnlyWS(temp)){
		  resize(nextIdx+1);
		  int offset = 0;
		  (*this)[nextIdx].mContigName = NextString(temp,offset);
		  (*this)[nextIdx].mSide = NextString(temp,offset);
		  (*this)[nextIdx].mStart = NextInt(temp,offset);
		  (*this)[nextIdx].mEnd = NextInt(temp,offset);

		  nextIdx++;
		}
	 }while(!fin.eof());
  }
};


void FindTail(vector<int> &TCover,int &leftTail, int &rightTail);
void FindCrash2(int leftTail,int rightTail, vector<int> &TCover,int &leftCrash, int &rightCrash);
void FindCrash(vector<int> &TCover,int &leftTail, int &rightTail);
void GetProblemContigs(string GraphFile,StringSet &ProblemContigs);

void IdentifySplits(Contig &C,ContigChecks &Checkss,vector<int> &leftSplits,
						  vector<int> &rightSplits);

#endif 
