#!/home/hgsc/bin/perl  

use Getopt::Long;
use File::Basename;
use Cwd;
use RCHEN::Scaffold;
use Bcm::LSF;
use Mail::Mailer;

# Assume done until proven otherwise.
$all_good_status="DONE";

$opt_root = "atlas/freeze04-stitch";
#$opt_root = "atlas/tuning16001";
$opt_delete = 1;

if (!GetOptions(
					 "help|h",
					 "file|f:s",  # The project list file.  
					 "root|r:s",  # The root directory for inputs and outputs. 
					 "delete|!"
					 )) {
    usage(1);
}

$output_root = $opt_root;


# Open up a global status file. 	 
$statname = "/users/pipeline/p-atlas/$output_root/status/$opt_file.stat";
print "Saving status to $statname\n";
open(STAT,">$statname") || die "Can't open $statname\n";
STAT->autoflush(1);
	 
$bactigID = $opt_file;	 

#main
{
	 my @all_ace=();	

	 $cleancmd ="rm -rf /scratch/core";
	 system($cleancmd);

	 $bactigpath = "/users/pipeline/p-atlas/$output_root/lists/$bactigID";
	 
    # Read in the paths of all of the ace files in this bactig
	 open(PROJ, $bactigpath) || die "Can not open the $bactigpath to read\n";
	 @all_ace = <PROJ>;
	 close PROJ; 

	 # make temporary directory
	 $tempdir = "/scratch/KJD$bactigID";

	 # hard remove it. 
	 $cleancmd ="rm -rf $tempdir";
	 system($cleancmd);

	 $cmd = "mkdir $tempdir";
	 print "$cmd";
	 RunCommand($cmd);

	 # Make a temp dir relative ace list. 
	 $tempacelist = "$tempdir/acelist";
	 open(TACELIST,">$tempacelist") || die "Can't open $tempacelist";

    # copy all of these ace files to the temporary directory. 
	 foreach $acefile (@all_ace) {
		  chomp($acefile);


		  # We're now looking at freeze00
		  #if ($acefile =~ "current"){
		#		$acefile =~ s/current/freeze00/;
		#  }		  

		  # parse out parts of acefile
		  if ($acefile =~ /\/\//){ 
				$acefile =~ s/\/\//\//;  # replace double / with single / 
		  }		  

		  if($acefile=~/projects\/([^\/]+)\//) {
				$dir_name= dirname($acefile);
				$project=$1;
		  } elsif($acefile=~/projects\/([^\/]+)$/) {
				$dir_name='/users/pipeline/p-atlas/projects/'.$1.'/current/';
				$acefile=$dir_name.'/atlas.fasta.ace.gz';
				$project=$1;
		  } else{
				$project=$acefile;
				$dir_name= '/users/pipeline/p-atlas/projects/'.$acefile.'/current/';
				$acefile =$dir_name.'/atlas.fasta.ace.gz';
		  }
		  
		  print "acefile=$acefile\n";
		  print "dir_name=$dir_name\n";
		  print "project=$project\n";

		  $qualityfile = $dir_name."/atlas.fasta.qual.gz";
		  		  
		  # copy ace file and quality. 
		  $newacename = "$project.ace.gz";
		  $cmd = "cp $acefile $tempdir/$newacename";
		  
		  RunCommand($cmd);

		  $newqualname = "$project.qual.gz";
		  $cmd = "cp $qualityfile $tempdir/$newqualname";
		  RunCommand($cmd);

		  print TACELIST "$tempdir/$newacename\n";
	 }
	 close(TACELIST);

	 # create a single big quality file
	 $cmd = "cat $tempdir/*qual.gz | gunzip > $tempdir/$bactigID.qual";
	 RunCommand($cmd);

	 # Run stitcher
	 $tilingpath = "/users/pipeline/p-atlas/atlas/freeze04-stitch/lists/$bactigID.path";
	 $outname = "$tempdir/$bactigID.ace";
	 $sout = "/users/pipeline/p-atlas/$output_root/stdout/$bactigID.stdout";
	 $cmd = "/home/hgsc/bin/atlas-simplestitcher-1.24 -t $tilingpath -f $tempacelist -o $outname -T $tempdir/TEMP > $sout 2>&1";
	 RunCommand($cmd);

	 $cmd = "gzip $outname $tempdir/$bactigID.qual";
	 RunCommand($cmd);
		
	 # Copy results 
	 $cmd = "cp -f $outname.gz /users/pipeline/p-atlas/$output_root/output/$bactigID/$bactigID.ace.gz";
	 RunCommand($cmd);

	 $cmd = "cp -f $tempdir/$bactigID.qual.gz /users/pipeline/p-atlas/$output_root/output/$bactigID/";
	 RunCommand($cmd);

	 # Check whether we ran out of space, tack results of df onto end of status
	 $cmd = "df $tempdir >> $statname\n";
	 RunCommand($cmd);


	if ($opt_delete){
			 # Clean up temporary directory. 
			 $cmd = "rm -rf $tempdir";
			 RunCommand($cmd);
	}

	 print STAT "$all_good_status\n";
	 close(STAT);
	 exit(1);
}	 




# Run a command, recording the command and it's status in the status file. 
sub RunCommand()
{
	 my $cmd = shift;
	 print STAT "$cmd\n";
	 $status = system($cmd);
	 
	 if ($status != 0) {
		  $all_good_status = "FAIL";
	 }
	 print STAT "status = $status\n";	 
}


sub usage()
{
		print"hi";
}

